"""
Vector search tool for retrieving documents from Databricks Vector Search.

This module provides a tool factory for creating semantic search tools
with dynamic filter schemas based on table columns, FlashRank reranking support,
instructed retrieval with query decomposition and RRF merging, and optional
query routing, result verification, and instruction-aware reranking.
"""

import json
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Annotated, Any, Literal, Optional

import mlflow
from databricks.sdk import WorkspaceClient
from databricks.vector_search.reranker import DatabricksReranker
from databricks_langchain import DatabricksVectorSearch
from flashrank import Ranker, RerankRequest
from langchain.tools import ToolRuntime, tool
from langchain_core.documents import Document
from langchain_core.tools import StructuredTool
from loguru import logger
from mlflow.entities import SpanType

from dao_ai.config import (
    ColumnInfo,
    DecompositionModel,
    FilterItem,
    InstructedRetrieverModel,
    InstructionAwareRerankModel,
    RerankParametersModel,
    RetrieverModel,
    RouterModel,
    SearchParametersModel,
    SearchQuery,
    VectorStoreModel,
    VerificationResult,
    VerifierModel,
    value_of,
)
from dao_ai.state import Context
from dao_ai.tools.instructed_retriever import (
    _get_cached_llm,
    decompose_query,
    rrf_merge,
)
from dao_ai.tools.instruction_reranker import instruction_aware_rerank
from dao_ai.tools.router import route_query
from dao_ai.tools.verifier import add_verification_metadata, verify_results
from dao_ai.utils import is_in_model_serving, normalize_host


@mlflow.trace(name="rerank_documents", span_type=SpanType.RERANKER)
def _rerank_documents(
    query: str,
    documents: list[Document],
    ranker: Ranker,
    rerank_config: RerankParametersModel,
) -> list[Document]:
    """
    Rerank documents using FlashRank cross-encoder model.

    Args:
        query: The search query string
        documents: List of documents to rerank
        ranker: The FlashRank Ranker instance
        rerank_config: Reranking configuration

    Returns:
        Reranked list of documents with reranker_score in metadata
    """
    logger.trace(
        "Starting reranking",
        documents_count=len(documents),
        model=rerank_config.model,
    )

    # Early return if no documents to rerank
    if not documents:
        logger.debug("No documents to rerank, skipping")
        return documents

    # Prepare passages for reranking
    passages: list[dict[str, Any]] = [
        {"text": doc.page_content, "meta": doc.metadata} for doc in documents
    ]

    # Create reranking request
    rerank_request: RerankRequest = RerankRequest(query=query, passages=passages)

    # Perform reranking
    results: list[dict[str, Any]] = ranker.rerank(rerank_request)

    # Apply top_n filtering
    top_n: int = rerank_config.top_n or len(documents)
    results = results[:top_n]
    logger.debug("Reranking complete", top_n=top_n, candidates_count=len(documents))

    # Convert back to Document objects with reranking scores
    reranked_docs: list[Document] = []
    for result in results:
        orig_doc: Optional[Document] = next(
            (doc for doc in documents if doc.page_content == result["text"]), None
        )
        if orig_doc:
            reranked_doc: Document = Document(
                page_content=orig_doc.page_content,
                metadata={
                    **orig_doc.metadata,
                    "reranker_score": result["score"],
                },
            )
            reranked_docs.append(reranked_doc)

    logger.debug(
        "Documents reranked",
        input_count=len(documents),
        output_count=len(reranked_docs),
        model=rerank_config.model,
    )

    return reranked_docs


def create_vector_search_tool(
    retriever: Optional[RetrieverModel | dict[str, Any]] = None,
    vector_store: Optional[VectorStoreModel | dict[str, Any]] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> StructuredTool:
    """
    Create a Vector Search tool with dynamic schema and optional reranking.

    Args:
        retriever: Full retriever configuration with search parameters and reranking
        vector_store: Direct vector store reference (uses default search parameters)
        name: Optional custom name for the tool
        description: Optional custom description for the tool

    Returns:
        A LangChain StructuredTool with proper schema (additionalProperties: false)
    """

    # Validate mutually exclusive parameters
    if retriever is None and vector_store is None:
        raise ValueError("Must provide either 'retriever' or 'vector_store' parameter")
    if retriever is not None and vector_store is not None:
        raise ValueError(
            "Cannot provide both 'retriever' and 'vector_store' parameters"
        )

    # Handle vector_store parameter
    if vector_store is not None:
        if isinstance(vector_store, dict):
            vector_store = VectorStoreModel(**vector_store)
        retriever = RetrieverModel(vector_store=vector_store)
    else:
        if isinstance(retriever, dict):
            retriever = RetrieverModel(**retriever)

    vector_store: VectorStoreModel = retriever.vector_store

    # Index is required
    if vector_store.index is None:
        raise ValueError("vector_store.index is required for vector search")

    index_name: str = vector_store.index.full_name
    columns: list[str] = list(retriever.columns or vector_store.index.columns or [])
    search_parameters: SearchParametersModel = retriever.search_parameters
    rerank_config: Optional[RerankParametersModel] = retriever.rerank
    instructed_config: Optional[InstructedRetrieverModel] = retriever.instructed

    # Extract nested configs from instructed (all depend on schema context)
    # Columns are the single source of truth for schema context.
    instructed_columns: list[ColumnInfo] = (
        instructed_config.columns if instructed_config else []
    )
    decomposition_config: Optional[DecompositionModel] = (
        instructed_config.decomposition if instructed_config else None
    )
    router_config: Optional[RouterModel] = (
        instructed_config.router if instructed_config else None
    )
    instruction_rerank_config: Optional[InstructionAwareRerankModel] = (
        instructed_config.rerank if instructed_config else None
    )
    verifier_config: Optional[VerifierModel] = (
        instructed_config.verifier if instructed_config else None
    )

    # Initialize FlashRank ranker if configured
    ranker: Optional[Ranker] = None
    if rerank_config and rerank_config.model:
        logger.debug(
            "Initializing FlashRank ranker",
            model=rerank_config.model,
            top_n=rerank_config.top_n or "auto",
        )
        try:
            # Use /tmp for cache in Model Serving (home dir may not be writable)
            if is_in_model_serving():
                cache_dir = "/tmp/dao_ai/cache/flashrank"
                if rerank_config.cache_dir != cache_dir:
                    logger.warning(
                        "FlashRank cache_dir overridden in Model Serving",
                        configured=rerank_config.cache_dir,
                        actual=cache_dir,
                    )
            else:
                cache_dir = os.path.expanduser(rerank_config.cache_dir)
            ranker = Ranker(model_name=rerank_config.model, cache_dir=cache_dir)

            # Patch rerank to always include token_type_ids for ONNX compatibility
            # Some ONNX runtimes require token_type_ids even when the model doesn't use them
            # FlashRank conditionally excludes them when all zeros, but ONNX may still expect them
            # See: https://github.com/huggingface/optimum/issues/1500
            if ranker.session is not None:
                import numpy as np

                _original_rerank = ranker.rerank

                def _patched_rerank(request):
                    query = request.query
                    passages = request.passages
                    query_passage_pairs = [[query, p["text"]] for p in passages]

                    input_text = ranker.tokenizer.encode_batch(query_passage_pairs)
                    input_ids = np.array([e.ids for e in input_text])
                    token_type_ids = np.array([e.type_ids for e in input_text])
                    attention_mask = np.array([e.attention_mask for e in input_text])

                    # Always include token_type_ids (the fix for ONNX compatibility)
                    onnx_input = {
                        "input_ids": input_ids.astype(np.int64),
                        "attention_mask": attention_mask.astype(np.int64),
                        "token_type_ids": token_type_ids.astype(np.int64),
                    }

                    outputs = ranker.session.run(None, onnx_input)
                    logits = outputs[0]

                    if logits.shape[1] == 1:
                        scores = 1 / (1 + np.exp(-logits.flatten()))
                    else:
                        exp_logits = np.exp(logits)
                        scores = exp_logits[:, 1] / np.sum(exp_logits, axis=1)

                    for score, passage in zip(scores, passages):
                        passage["score"] = score

                    passages.sort(key=lambda x: x["score"], reverse=True)
                    return passages

                ranker.rerank = _patched_rerank

            logger.success("FlashRank ranker initialized", model=rerank_config.model)
        except Exception as e:
            logger.warning("Failed to initialize FlashRank ranker", error=str(e))
            rerank_config = None

    # Log instructed retrieval configuration
    if instructed_config and decomposition_config:
        logger.success(
            "Instructed retrieval configured",
            decomposition_model=decomposition_config.model.name
            if decomposition_config.model
            else None,
            max_subqueries=decomposition_config.max_subqueries,
            rrf_k=decomposition_config.rrf_k,
        )

    # Log instruction-aware reranking configuration
    if instruction_rerank_config:
        logger.success(
            "Instruction-aware reranking configured",
            model=instruction_rerank_config.model.name
            if instruction_rerank_config.model
            else None,
            top_n=instruction_rerank_config.top_n,
        )

    # Build client_args for VectorSearchClient
    client_args: dict[str, Any] = {}
    has_explicit_auth = any(
        [
            os.environ.get("DATABRICKS_TOKEN"),
            os.environ.get("DATABRICKS_CLIENT_ID"),
            vector_store.pat,
            vector_store.client_id,
            vector_store.on_behalf_of_user,
        ]
    )

    if has_explicit_auth:
        databricks_host = os.environ.get("DATABRICKS_HOST")
        if not databricks_host and vector_store.workspace_host:
            databricks_host = value_of(vector_store.workspace_host)
        if databricks_host:
            client_args["workspace_url"] = normalize_host(databricks_host)

        token = os.environ.get("DATABRICKS_TOKEN")
        if not token and vector_store.pat:
            token = value_of(vector_store.pat)
        if token:
            client_args["personal_access_token"] = token

        client_id = os.environ.get("DATABRICKS_CLIENT_ID")
        if not client_id and vector_store.client_id:
            client_id = value_of(vector_store.client_id)
        if client_id:
            client_args["service_principal_client_id"] = client_id

        client_secret = os.environ.get("DATABRICKS_CLIENT_SECRET")
        if not client_secret and vector_store.client_secret:
            client_secret = value_of(vector_store.client_secret)
        if client_secret:
            client_args["service_principal_client_secret"] = client_secret

    logger.debug(
        "Creating vector search tool",
        name=name,
        index=index_name,
        client_args_keys=list(client_args.keys()) if client_args else [],
    )

    # Cache for DatabricksVectorSearch - created lazily for OBO support
    _cached_vector_search: DatabricksVectorSearch | None = None

    def _get_vector_search(context: Context | None) -> DatabricksVectorSearch:
        """Get or create DatabricksVectorSearch, using context for OBO auth if available."""
        nonlocal _cached_vector_search

        # Use cached instance if available and not OBO
        if _cached_vector_search is not None and not vector_store.on_behalf_of_user:
            return _cached_vector_search

        # Get workspace client with OBO support via context
        workspace_client: WorkspaceClient = vector_store.workspace_client_from(context)

        # Create DatabricksVectorSearch
        # Note: text_column should be None for Databricks-managed embeddings
        # (it's automatically determined from the index)
        vs: DatabricksVectorSearch = DatabricksVectorSearch(
            index_name=index_name,
            text_column=None,
            columns=columns,
            workspace_client=workspace_client,
            client_args=client_args if client_args else None,
            primary_key=vector_store.primary_key,
            doc_uri=vector_store.doc_uri,
            include_score=True,
            reranker=(
                DatabricksReranker(columns_to_rerank=rerank_config.columns)
                if rerank_config and rerank_config.columns
                else None
            ),
        )

        # Cache for non-OBO scenarios
        if not vector_store.on_behalf_of_user:
            _cached_vector_search = vs

        return vs

    # Determine tool name and description
    tool_name: str = name or f"vector_search_{vector_store.index.name}"

    # Build tool description with available columns for filtering
    base_description: str = description or f"Search documents in {index_name}"
    if columns:
        columns_list = ", ".join(columns)
        tool_description = (
            f"{base_description}. "
            f"Available filter columns: {columns_list}. "
            f"Filter operators: 'column' for equality, 'column NOT' for exclusion, "
            f"'column <', 'column <=', 'column >', 'column >=' for comparison, "
            f"'column LIKE' for token matching, 'column NOT LIKE' to exclude tokens."
        )
    else:
        tool_description = base_description

    @mlflow.trace(name="execute_instructed_retrieval", span_type=SpanType.RETRIEVER)
    def _execute_instructed_retrieval(
        vs: DatabricksVectorSearch,
        query: str,
        base_filters: dict[str, Any],
        previous_feedback: str | None = None,
    ) -> list[Document]:
        """Execute instructed retrieval with query decomposition and RRF merging."""
        logger.trace(
            "Executing instructed retrieval", query=query, base_filters=base_filters
        )
        try:
            decomposition_llm = _get_cached_llm(decomposition_config.model)

            subqueries: list[SearchQuery] = decompose_query(
                llm=decomposition_llm,
                query=query,
                columns=instructed_columns,
                constraints=instructed_config.constraints,
                max_subqueries=decomposition_config.max_subqueries,
                examples=decomposition_config.examples,
                previous_feedback=previous_feedback,
            )

            if not subqueries:
                logger.warning(
                    "Query decomposition returned no subqueries, using original"
                )
                return vs.similarity_search(
                    query=query,
                    k=search_parameters.num_results or 5,
                    filter=base_filters if base_filters else None,
                    query_type=search_parameters.query_type or "ANN",
                )

            def normalize_filter_values(
                filters: dict[str, Any], case: str | None
            ) -> dict[str, Any]:
                """Normalize string filter values to specified case."""
                logger.trace("Normalizing filter values", filters=filters, case=case)
                if not case or not filters:
                    return filters
                normalized = {}
                for key, value in filters.items():
                    if isinstance(value, str):
                        normalized[key] = (
                            value.upper() if case == "uppercase" else value.lower()
                        )
                    elif isinstance(value, list):
                        normalized[key] = [
                            v.upper()
                            if case == "uppercase"
                            else v.lower()
                            if isinstance(v, str)
                            else v
                            for v in value
                        ]
                    else:
                        normalized[key] = value
                return normalized

            def execute_search(sq: SearchQuery) -> list[Document]:
                logger.trace("Executing search", query=sq.text, filters=sq.filters)
                # Convert FilterItem list to dict
                sq_filters_dict: dict[str, Any] = {}
                if sq.filters:
                    for item in sq.filters:
                        sq_filters_dict[item.key] = item.value
                sq_filters = normalize_filter_values(
                    sq_filters_dict, decomposition_config.normalize_filter_case
                )
                k: int = search_parameters.num_results or 5
                query_type: str = search_parameters.query_type or "ANN"
                combined_filters: dict[str, Any] = {**sq_filters, **base_filters}
                logger.trace(
                    "Executing search",
                    query=sq.text,
                    k=k,
                    query_type=query_type,
                    filters=combined_filters,
                )
                return vs.similarity_search(
                    query=sq.text,
                    k=k,
                    filter=combined_filters if combined_filters else None,
                    query_type=query_type,
                )

            logger.debug(
                "Executing parallel searches",
                num_subqueries=len(subqueries),
                queries=[sq.text[:50] for sq in subqueries],
            )

            with ThreadPoolExecutor(
                max_workers=decomposition_config.max_subqueries
            ) as executor:
                all_results = list(executor.map(execute_search, subqueries))

            merged = rrf_merge(
                all_results,
                k=decomposition_config.rrf_k,
                primary_key=vector_store.primary_key,
            )

            logger.debug(
                "Instructed retrieval complete",
                num_subqueries=len(subqueries),
                total_results=sum(len(r) for r in all_results),
                merged_results=len(merged),
            )

            return merged

        except Exception as e:
            logger.warning(
                "Instructed retrieval failed, falling back to standard search",
                error=str(e),
            )
            return vs.similarity_search(
                query=query,
                k=search_parameters.num_results or 5,
                filter=base_filters if base_filters else None,
                query_type=search_parameters.query_type or "ANN",
            )

    @mlflow.trace(name="execute_standard_search", span_type=SpanType.RETRIEVER)
    def _execute_standard_search(
        vs: DatabricksVectorSearch,
        query: str,
        base_filters: dict[str, Any],
    ) -> list[Document]:
        """Execute standard single-query search."""
        logger.trace("Performing standard vector search", query_preview=query[:50])
        return vs.similarity_search(
            query=query,
            k=search_parameters.num_results or 5,
            filter=base_filters if base_filters else None,
            query_type=search_parameters.query_type or "ANN",
        )

    @mlflow.trace(name="apply_post_processing", span_type=SpanType.RETRIEVER)
    def _apply_post_processing(
        documents: list[Document],
        query: str,
        mode: Literal["standard", "instructed"],
        auto_bypass: bool,
    ) -> list[Document]:
        """Apply instruction-aware reranking and verification based on mode and bypass settings."""
        # Skip post-processing for standard mode when auto_bypass is enabled
        if mode == "standard" and auto_bypass:
            mlflow.set_tag("router.bypassed_stages", "true")
            return documents

        # Apply instruction-aware reranking if configured
        if instruction_rerank_config and instructed_config:
            instruction_llm = (
                _get_cached_llm(instruction_rerank_config.model)
                if instruction_rerank_config.model
                else None
            )

            if instruction_llm:
                documents = instruction_aware_rerank(
                    llm=instruction_llm,
                    query=query,
                    documents=documents,
                    instructions=instruction_rerank_config.instructions,
                    columns=instructed_columns,
                    top_n=instruction_rerank_config.top_n,
                )

        # Apply verification if configured (verifier is always under instructed)
        if verifier_config and instructed_config:
            verifier_llm = (
                _get_cached_llm(verifier_config.model)
                if verifier_config.model
                else None
            )

            if verifier_llm:
                constraints = instructed_config.constraints
                retry_count = 0
                verification_result: VerificationResult | None = None
                previous_feedback: str | None = None

                while retry_count <= verifier_config.max_retries:
                    verification_result = verify_results(
                        llm=verifier_llm,
                        query=query,
                        documents=documents,
                        columns=instructed_columns,
                        constraints=constraints,
                        previous_feedback=previous_feedback,
                    )

                    if verification_result.passed:
                        mlflow.set_tag("verifier.outcome", "passed")
                        mlflow.set_tag("verifier.retries", str(retry_count))
                        break

                    # Handle failure based on configuration
                    if verifier_config.on_failure == "warn":
                        mlflow.set_tag("verifier.outcome", "warned")
                        documents = add_verification_metadata(
                            documents, verification_result
                        )
                        break

                    if retry_count >= verifier_config.max_retries:
                        mlflow.set_tag("verifier.outcome", "exhausted")
                        mlflow.set_tag("verifier.retries", str(retry_count))
                        documents = add_verification_metadata(
                            documents, verification_result, exhausted=True
                        )
                        break

                    # Retry with feedback
                    mlflow.set_tag("verifier.outcome", "retried")
                    previous_feedback = verification_result.feedback
                    retry_count += 1
                    logger.debug(
                        "Retrying search with verification feedback", retry=retry_count
                    )

        return documents

    # Use @tool decorator for proper ToolRuntime injection
    @tool(name_or_callable=tool_name, description=tool_description)
    def _vector_search_tool(
        query: Annotated[str, "The search query to find relevant documents"],
        filters: Annotated[
            Optional[list[FilterItem]],
            "Optional filters as key-value pairs. "
            "Key operators: 'column' (equality), 'column NOT' (exclusion), "
            "'column <', '<=', '>', '>=' (comparison), "
            "'column LIKE' (token match), 'column NOT LIKE' (exclude token). "
            f"Valid columns: {', '.join(columns) if columns else 'none'}.",
        ] = None,
        runtime: ToolRuntime[Context] = None,
    ) -> str:
        """Search for relevant documents using vector similarity."""
        context: Context | None = runtime.context if runtime else None
        vs: DatabricksVectorSearch = _get_vector_search(context)

        filters_dict: dict[str, Any] = {}
        if filters:
            for item in filters:
                filters_dict[item.key] = item.value

        base_filters: dict[str, Any] = {
            **filters_dict,
            **(search_parameters.filters or {}),
        }

        # Determine execution mode via router or config
        mode: Literal["standard", "instructed"] = "standard"
        auto_bypass = True

        logger.trace("Router configuration", router_config=router_config)
        logger.trace("Instructed configuration", instructed_config=instructed_config)
        logger.trace(
            "Instruction-aware rerank configuration",
            instruction_aware=instruction_rerank_config,
        )

        if router_config:
            router_llm = (
                _get_cached_llm(router_config.model) if router_config.model else None
            )
            auto_bypass = router_config.auto_bypass

            if router_llm and instructed_config:
                try:
                    mode = route_query(
                        llm=router_llm,
                        query=query,
                        columns=instructed_columns,
                    )
                except Exception as e:
                    # Router fail-safe: default to standard mode
                    logger.warning(
                        "Router failed, defaulting to standard mode", error=str(e)
                    )
                    mlflow.set_tag("router.fallback", "true")
                    mode = router_config.default_mode
            else:
                mode = router_config.default_mode
        elif instructed_config:
            # No router but instructed is configured - use instructed mode
            mode = "instructed"
            auto_bypass = False

        logger.trace("Routing mode", mode=mode, auto_bypass=auto_bypass)
        mlflow.set_tag("router.mode", mode)

        # Execute search based on mode
        if mode == "instructed" and instructed_config and decomposition_config:
            documents = _execute_instructed_retrieval(vs, query, base_filters)
        else:
            documents = _execute_standard_search(vs, query, base_filters)

        # Apply FlashRank reranking if configured
        if ranker and rerank_config:
            logger.debug("Applying FlashRank reranking")
            documents = _rerank_documents(query, documents, ranker, rerank_config)

        # Apply post-processing (instruction reranking + verification)
        documents = _apply_post_processing(documents, query, mode, auto_bypass)

        # Serialize documents to JSON format for LLM consumption
        serialized_docs: list[dict[str, Any]] = []
        for doc in documents:
            metadata_serializable: dict[str, Any] = {}
            for key, value in doc.metadata.items():
                if hasattr(value, "item"):  # numpy scalar
                    metadata_serializable[key] = value.item()
                else:
                    metadata_serializable[key] = value

            serialized_docs.append(
                {
                    "page_content": doc.page_content,
                    "metadata": metadata_serializable,
                }
            )

        return json.dumps(serialized_docs)

    logger.success("Vector search tool created", name=tool_name, index=index_name)

    return _vector_search_tool



"""
Tests for tool retry middleware factory.
"""

import pytest
from langchain.agents.middleware import ToolRetryMiddleware

from dao_ai.config import PythonFunctionModel, ToolModel
from dao_ai.middleware import create_tool_retry_middleware


class TestCreateToolRetryMiddleware:
    """Tests for the create_tool_retry_middleware factory function."""

    def test_create_with_defaults(self):
        """Test creating middleware with default parameters."""
        middleware = create_tool_retry_middleware()

        # Middleware is single instance
        assert middleware is not None
        middleware = middleware
        assert isinstance(middleware, ToolRetryMiddleware)
        assert middleware.max_retries == 3
        assert middleware.backoff_factor == 2.0
        assert middleware.initial_delay == 1.0
        assert middleware.on_failure == "continue"

    def test_create_with_custom_retries(self):
        """Test creating middleware with custom max_retries."""
        middleware = create_tool_retry_middleware(max_retries=5)

        assert middleware.max_retries == 5

    def test_create_with_custom_backoff(self):
        """Test creating middleware with custom backoff settings."""
        middleware = create_tool_retry_middleware(
            backoff_factor=1.5,
            initial_delay=0.5,
        )

        middleware = middleware
        assert middleware.backoff_factor == 1.5
        assert middleware.initial_delay == 0.5

    def test_create_with_max_delay(self):
        """Test creating middleware with max_delay cap."""
        middleware = create_tool_retry_middleware(
            max_retries=10,
            max_delay=60.0,
        )

        assert middleware.max_delay == 60.0

    def test_create_with_jitter(self):
        """Test creating middleware with jitter enabled."""
        middleware = create_tool_retry_middleware(jitter=True)

        assert middleware.jitter is True

    def test_create_with_string_tools(self):
        """Test creating middleware for specific tools by name."""
        middleware = create_tool_retry_middleware(
            tools=["search_web", "query_database"],
        )

        # Middleware should be created successfully with tool filtering
        # Middleware is single instance
        assert middleware is not None
        assert isinstance(middleware, ToolRetryMiddleware)

    def test_create_with_tool_model(self):
        """Test creating middleware with ToolModel."""
        tool_model = ToolModel(
            name="my_tool",
            function=PythonFunctionModel(name="dao_ai.tools.say_hello_tool"),
        )

        middleware = create_tool_retry_middleware(tools=[tool_model])

        # Should create middleware successfully
        # Middleware is single instance
        assert middleware is not None
        assert isinstance(middleware, ToolRetryMiddleware)

    def test_create_with_dict_tool(self):
        """Test creating middleware with dict tool config."""
        tool_dict = {
            "name": "my_tool",
            "function": {"name": "dao_ai.tools.say_hello_tool"},
        }

        middleware = create_tool_retry_middleware(tools=[tool_dict])

        # Middleware is single instance
        assert middleware is not None

    def test_create_with_error_on_failure(self):
        """Test creating middleware with error on_failure behavior."""
        middleware = create_tool_retry_middleware(on_failure="error")

        assert middleware.on_failure == "error"

    def test_create_with_exception_tuple(self):
        """Test creating middleware with specific exception types."""
        middleware = create_tool_retry_middleware(
            retry_on=(TimeoutError, ConnectionError),
        )

        assert middleware.retry_on == (TimeoutError, ConnectionError)

    def test_create_with_retry_callable(self):
        """Test creating middleware with callable retry_on."""

        def should_retry(error: Exception) -> bool:
            return "rate_limit" in str(error).lower()

        middleware = create_tool_retry_middleware(retry_on=should_retry)

        assert middleware.retry_on is should_retry

    def test_create_with_custom_on_failure_callable(self):
        """Test creating middleware with callable on_failure."""

        def format_error(error: Exception) -> str:
            return f"Tool failed: {error}"

        middleware = create_tool_retry_middleware(on_failure=format_error)

        assert middleware.on_failure is format_error

    def test_create_with_invalid_dict(self):
        """Test that invalid dict raises helpful error."""
        invalid_dict = {"invalid": "data"}

        with pytest.raises(ValueError, match="Failed to construct ToolModel from dict"):
            create_tool_retry_middleware(tools=[invalid_dict])

    def test_create_all_tools_when_none(self):
        """Test that None tools applies to all tools."""
        middleware = create_tool_retry_middleware(tools=None)

        # Should create successfully without tool filtering
        # Middleware is single instance
        assert middleware is not None
        assert isinstance(middleware, ToolRetryMiddleware)

    def test_returns_list_for_composition(self):
        """Test that factory returns list for easy composition."""
        middleware = create_tool_retry_middleware()

        # Middleware is single instance
        assert middleware is not None



import importlib
import importlib.metadata
import json
import os
import re
import site
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import Any, Callable, Sequence, TypeVar

from langchain_core.language_models import BaseChatModel
from langchain_core.tools import BaseTool
from loguru import logger
from pydantic import BaseModel

import dao_ai

T = TypeVar("T", bound=BaseModel)


def is_lib_provided(lib_name: str, pip_requirements: Sequence[str]) -> bool:
    return any(
        re.search(rf"\b{re.escape(lib_name)}\b", requirement)
        for requirement in pip_requirements
    )


def is_installed() -> bool:
    current_file = os.path.abspath(dao_ai.__file__)
    site_packages = [os.path.abspath(path) for path in site.getsitepackages()]
    if site.getusersitepackages():
        site_packages.append(os.path.abspath(site.getusersitepackages()))

    found: bool = any(current_file.startswith(pkg_path) for pkg_path in site_packages)
    logger.trace(
        "Checking if dao_ai is installed", is_installed=found, current_file=current_file
    )
    return found


def normalize_name(name: str) -> str:
    normalized = re.sub(r"[^a-zA-Z0-9_]", "_", name).lower()
    normalized = re.sub(r"_+", "_", normalized)
    return normalized.strip("_")


def normalize_host(host: str | None) -> str | None:
    """Ensure host URL has https:// scheme.

    The DATABRICKS_HOST environment variable should always include the https://
    scheme, but some environments (e.g., Databricks Apps infrastructure) may
    provide the host without it. This function normalizes the host to ensure
    it has the proper scheme.

    Args:
        host: The host URL, with or without scheme

    Returns:
        The host URL with https:// scheme, or None if host is None/empty
    """
    if not host:
        return None
    host = host.strip()
    if not host:
        return None
    if not host.startswith("http://") and not host.startswith("https://"):
        return f"https://{host}"
    return host


def get_default_databricks_host() -> str | None:
    """Get the default Databricks workspace host.

    Attempts to get the host from:
    1. DATABRICKS_HOST environment variable
    2. WorkspaceClient ambient authentication (e.g., from ~/.databrickscfg)

    Returns:
        The Databricks workspace host URL (with https:// scheme), or None if not available.
    """
    # Try environment variable first
    host: str | None = os.environ.get("DATABRICKS_HOST")
    if host:
        return normalize_host(host)

    # Fall back to WorkspaceClient
    try:
        from databricks.sdk import WorkspaceClient

        w: WorkspaceClient = WorkspaceClient()
        return normalize_host(w.config.host)
    except Exception:
        logger.trace("Could not get default Databricks host from WorkspaceClient")
        return None


def dao_ai_version() -> str:
    """
    Get the dao-ai package version, with fallback for source installations.

    Tries to get the version from installed package metadata first. If the package
    is not installed (e.g., running from source), falls back to reading from
    pyproject.toml. Returns "dev" if neither method works.

    Returns:
        str: The version string, or "dev" if version cannot be determined
    """
    try:
        # Try to get version from installed package metadata
        return version("dao-ai")
    except PackageNotFoundError:
        # Package not installed, try reading from pyproject.toml
        logger.trace(
            "dao-ai package not installed, attempting to read version from pyproject.toml"
        )
        try:
            import tomllib  # Python 3.11+
        except ImportError:
            try:
                import tomli as tomllib  # Fallback for Python < 3.11
            except ImportError:
                logger.warning(
                    "Cannot determine dao-ai version: package not installed and tomllib/tomli not available"
                )
                return "dev"

        try:
            # Find pyproject.toml relative to this file
            project_root = Path(__file__).parents[2]
            pyproject_path = project_root / "pyproject.toml"

            if not pyproject_path.exists():
                logger.warning(
                    "Cannot determine dao-ai version: pyproject.toml not found",
                    path=str(pyproject_path),
                )
                return "dev"

            with open(pyproject_path, "rb") as f:
                pyproject_data = tomllib.load(f)
                pkg_version = pyproject_data.get("project", {}).get("version", "dev")
                logger.trace(
                    "Read version from pyproject.toml",
                    version=pkg_version,
                    path=str(pyproject_path),
                )
                return pkg_version
        except Exception as e:
            logger.warning(
                "Cannot determine dao-ai version from pyproject.toml", error=str(e)
            )
            return "dev"


def get_installed_packages() -> list[str]:
    """Get all installed packages with versions.

    Returns a list of pip requirement strings for packages used by dao-ai.
    This is used for MLflow model logging to ensure all dependencies are captured.
    """

    packages: list[str] = [
        f"databricks-agents=={version('databricks-agents')}",
        f"databricks-langchain[memory]=={version('databricks-langchain')}",
        f"databricks-mcp=={version('databricks-mcp')}",
        f"databricks-sdk[openai]=={version('databricks-sdk')}",
        f"ddgs=={version('ddgs')}",
        f"flashrank=={version('flashrank')}",
        f"langchain=={version('langchain')}",
        f"langchain-mcp-adapters=={version('langchain-mcp-adapters')}",
        f"langchain-openai=={version('langchain-openai')}",
        f"langchain-tavily=={version('langchain-tavily')}",
        f"langgraph=={version('langgraph')}",
        f"langgraph-checkpoint-postgres=={version('langgraph-checkpoint-postgres')}",
        f"langmem=={version('langmem')}",
        f"loguru=={version('loguru')}",
        f"mcp=={version('mcp')}",
        f"mlflow=={version('mlflow')}",
        f"nest-asyncio=={version('nest-asyncio')}",
        f"openevals=={version('openevals')}",
        f"openpyxl=={version('openpyxl')}",
        f"psycopg[binary,pool]=={version('psycopg')}",
        f"pydantic=={version('pydantic')}",
        f"pyyaml=={version('pyyaml')}",
        f"tomli=={version('tomli')}",
        f"unitycatalog-ai[databricks]=={version('unitycatalog-ai')}",
        f"unitycatalog-langchain[databricks]=={version('unitycatalog-langchain')}",
    ]
    return packages


def load_function(function_name: str) -> Callable[..., Any]:
    """
    Dynamically import and return a callable function using its fully qualified name.

    This utility function allows dynamic loading of functions from their string
    representation, enabling configuration-driven function resolution at runtime.
    It's particularly useful for loading different components based on configuration
    without hardcoding import statements.

    Args:
        fqn: Fully qualified name of the function to import, in the format
             "module.submodule.function_name"

    Returns:
        The imported callable function or langchain tool

    Raises:
        ImportError: If the module cannot be imported
        AttributeError: If the function doesn't exist in the module
        TypeError: If the resolved object is not callable or invocable

    Example:
        >>> func = callable_from_fqn("dao_ai.models.get_latest_model_version")
        >>> version = func("my_model")
    """
    logger.trace("Loading function", function_name=function_name)

    try:
        # Split the FQN into module path and function name
        module_path, func_name = function_name.rsplit(".", 1)

        # Dynamically import the module
        module = importlib.import_module(module_path)

        # Get the function from the module
        func: Any = getattr(module, func_name)

        # Verify that the resolved object is callable or is a LangChain tool
        # In langchain 1.x, StructuredTool objects are not directly callable
        # but have an invoke() method
        is_callable: bool = callable(func)
        is_langchain_tool: bool = isinstance(func, BaseTool)

        if not is_callable and not is_langchain_tool:
            raise TypeError(f"Function {func_name} is not callable or invocable.")

        return func
    except (ImportError, AttributeError, TypeError) as e:
        # Provide a detailed error message that includes the original exception
        raise ImportError(f"Failed to import {function_name}: {e}")


def type_from_fqn(type_name: str) -> type:
    """
    Load a type from a fully qualified name (FQN).

    Dynamically imports and returns a type (class) from a module using its
    fully qualified name. Useful for loading Pydantic models, dataclasses,
    or any Python type specified as a string in configuration files.

    Args:
        type_name: Fully qualified type name in format "module.path.ClassName"

    Returns:
        The imported type/class

    Raises:
        ValueError: If the FQN format is invalid
        ImportError: If the module cannot be imported
        AttributeError: If the type doesn't exist in the module
        TypeError: If the resolved object is not a type

    Example:
        >>> ProductModel = type_from_fqn("my_models.ProductInfo")
        >>> instance = ProductModel(name="Widget", price=9.99)
    """
    logger.trace("Loading type", type_name=type_name)

    try:
        # Split the FQN into module path and class name
        parts = type_name.rsplit(".", 1)
        if len(parts) != 2:
            raise ValueError(
                f"Invalid type name '{type_name}'. "
                "Expected format: 'module.path.ClassName'"
            )

        module_path, class_name = parts

        # Dynamically import the module
        try:
            module = importlib.import_module(module_path)
        except ModuleNotFoundError as e:
            raise ImportError(
                f"Could not import module '{module_path}' for type '{type_name}': {e}"
            ) from e

        # Get the class from the module
        if not hasattr(module, class_name):
            raise AttributeError(
                f"Module '{module_path}' does not have attribute '{class_name}'"
            )

        resolved_type = getattr(module, class_name)

        # Verify it's actually a type
        if not isinstance(resolved_type, type):
            raise TypeError(
                f"'{type_name}' resolved to {resolved_type}, which is not a type"
            )

        return resolved_type

    except (ValueError, ImportError, AttributeError, TypeError) as e:
        # Provide a detailed error message that includes the original exception
        raise type(e)(f"Failed to load type '{type_name}': {e}") from e


def is_in_model_serving() -> bool:
    """Check if running in Databricks Model Serving environment.

    Detects Model Serving by checking for environment variables that are
    typically set in that environment.
    """
    # Primary check - explicit Databricks Model Serving env var
    if os.environ.get("IS_IN_DB_MODEL_SERVING_ENV", "false").lower() == "true":
        return True

    # Secondary check - Model Serving sets these environment variables
    if os.environ.get("DATABRICKS_MODEL_SERVING_ENV"):
        return True

    # Check for cluster type indicator
    cluster_type = os.environ.get("DATABRICKS_CLUSTER_TYPE", "")
    if "model-serving" in cluster_type.lower():
        return True

    # Check for model serving specific paths
    if os.path.exists("/opt/conda/envs/mlflow-env"):
        return True

    return False


def get_databricks_response_format(model_class: type[BaseModel]) -> dict[str, Any]:
    """Create a Databricks-compatible response_format for structured output.

    Databricks requires the json_schema response format to have a 'name' field.
    This function creates the properly formatted response_format dictionary
    from a Pydantic model.

    Args:
        model_class: A Pydantic model class to use as the output schema

    Returns:
        A dictionary suitable for use with llm.bind(response_format=...)

    Example:
        >>> response_format = get_databricks_response_format(MyModel)
        >>> bound_llm = llm.bind(response_format=response_format)
        >>> result = bound_llm.invoke(prompt)
    """
    schema = model_class.model_json_schema()

    # Remove $defs from the schema - Databricks doesn't support complex refs
    # We need to inline any referenced definitions
    if "$defs" in schema:
        schema = _inline_schema_defs(schema)

    return {
        "type": "json_schema",
        "json_schema": {
            "name": model_class.__name__,
            "schema": schema,
            "strict": True,
        },
    }


def _inline_schema_defs(schema: dict[str, Any]) -> dict[str, Any]:
    """Inline $defs references in a JSON schema.

    Databricks doesn't support $ref and complex nested definitions,
    so we need to inline them.

    Args:
        schema: The original JSON schema with $defs

    Returns:
        A schema with all references inlined
    """
    defs = schema.pop("$defs", {})
    if not defs:
        return schema

    def resolve_refs(obj: Any) -> Any:
        if isinstance(obj, dict):
            if "$ref" in obj:
                # Extract the definition name from #/$defs/DefinitionName
                ref_path = obj["$ref"]
                if ref_path.startswith("#/$defs/"):
                    def_name = ref_path[len("#/$defs/") :]
                    if def_name in defs:
                        # Return a copy of the definition with refs resolved
                        return resolve_refs(defs[def_name].copy())
                return obj
            return {k: resolve_refs(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [resolve_refs(item) for item in obj]
        return obj

    return resolve_refs(schema)


def _repair_json(content: str) -> str | None:
    """Attempt to repair malformed JSON from LLM output.

    Handles common issues:
    - Extra text before/after JSON object
    - Truncated JSON (unclosed brackets/braces)
    - Trailing commas

    Args:
        content: The potentially malformed JSON string

    Returns:
        Repaired JSON string if successful, None otherwise
    """
    # 1. Extract JSON object if wrapped in extra text
    start = content.find("{")
    end = content.rfind("}")
    if start == -1 or end == -1 or start >= end:
        return None
    content = content[start : end + 1]

    # 2. Try parsing as-is first
    try:
        json.loads(content)
        return content
    except json.JSONDecodeError:
        pass

    # 3. Fix trailing commas before closing brackets
    content = re.sub(r",\s*}", "}", content)
    content = re.sub(r",\s*]", "]", content)

    # 4. Try to close unclosed brackets/braces
    open_braces = content.count("{") - content.count("}")
    open_brackets = content.count("[") - content.count("]")

    if open_braces > 0 or open_brackets > 0:
        # Remove trailing comma if present
        content = content.rstrip().rstrip(",")
        content += "]" * open_brackets + "}" * open_braces

    # 5. Final validation
    try:
        json.loads(content)
        return content
    except json.JSONDecodeError:
        return None


def invoke_with_structured_output(
    llm: BaseChatModel,
    prompt: str,
    model_class: type[T],
) -> T | None:
    """Invoke an LLM with Databricks-compatible structured output.

    Uses response_format with json_schema type and proper 'name' field
    as required by Databricks Foundation Model APIs.

    Args:
        llm: The language model to invoke
        prompt: The prompt to send to the model
        model_class: The Pydantic model class for the expected output

    Returns:
        An instance of model_class, or None if parsing fails
    """
    response_format = get_databricks_response_format(model_class)
    bound_llm = llm.bind(response_format=response_format)

    response = bound_llm.invoke(prompt)

    content = response.content
    if not isinstance(content, str):
        return None

    try:
        # Try parsing the JSON directly
        result_dict = json.loads(content)
        return model_class.model_validate(result_dict)
    except json.JSONDecodeError as e:
        # Attempt JSON repair
        repaired = _repair_json(content)
        if repaired:
            try:
                result_dict = json.loads(repaired)
                logger.debug("JSON repair successful", model_class=model_class.__name__)
                return model_class.model_validate(result_dict)
            except (json.JSONDecodeError, Exception):
                pass
        logger.warning(
            "Failed to parse structured output",
            error=str(e),
            model_class=model_class.__name__,
        )
        return None
    except Exception as e:
        logger.warning(
            "Failed to parse structured output",
            error=str(e),
            model_class=model_class.__name__,
        )
        return None
